@@ -1,7 +1,39 @@
+/*  $Id$
+
+    Part of SWI-Prolog
+
+    Author:        Jan Wielemaker
+    E-mail:        wielemak@science.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2007, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
+*/
+
 :- module(settings,
-	  [ setting/4,			% :Name, +Type, +Default, +Comment, +Src
+	  [ setting/4,			% :Name, +Type, +Default, +Comment
 	    setting/2,			% :Name, ?Value
 	    set_setting/2,		% :Name, +Value
+	    set_setting_default/2,	% :Name, +Value
 	    restore_setting/1,		% :Name
 	    load_settings/1,		% +File
 	    load_settings/2,		% +File, +Options
@@ -48,6 +80,7 @@ access, loading and saving of settings.
 
 :- dynamic
 	value/3,			% Name, Module, Value
+	default/3,			% Name, Module, Value
 	local_file/1.			% Path
 
 :- multifile
@@ -57,9 +90,17 @@ access, loading and saving of settings.
 	setting(:, +, +, +),
 	setting(:, ?),
 	set_setting(:, +),
+	set_setting_default(:, +),
 	current_setting(:),
 	restore_setting(:).
 
+curr_setting(Name, Module, Type, Default, Comment) :-
+	current_setting(Name, Module, Type, Default0, Comment, _Src),
+	(   default(Name, Module, Default1)
+	->  Default = Default1
+	;   Default = Default0
+	).
+
 %%	setting(Name, Type, Default, Comment) is det.
 %
 %	Define a setting. Name denotes the name of the setting, Type its
@@ -91,7 +132,7 @@ user:term_expansion((:- setting(QName, Type, Default, Comment)),
 	(   current_setting(Name, Module, _, _, _, OldLoc)
 	->  format(string(Message),
 		   'Already defined at: ~w', [OldLoc]),
-	    throw(error(permission_error(redefine, setting, Name),
+	    throw(error(permission_error(redefine, setting, Module:Name),
 			context(Message, _)))
 	;   source_location(File, Line)
 	->  Expanded = settings:current_setting(Name, Module, Type, Default,
@@ -114,7 +155,7 @@ setting(QName, Value) :-
 	(   ground(Name)
 	->  (   value(Name, Module, Value0)
 	    ->  Value = Value0
-	    ;   current_setting(Name, Module, Type, Default, _, _)
+	    ;   curr_setting(Name, Module, Type, Default, _)
 	    ->	eval_default(Default, Module, Type, Value)
 	    ;	existence_error(setting, Module:Name)
 	    )
@@ -288,7 +329,7 @@ numeric_type(between(L,_), Type) :-
 set_setting(QName, Value) :-
 	strip_module(QName, Module, Name),
 	must_be(atom, Name),
-	(   current_setting(Name, Module, Type, Default0, _Comment, _Src),
+	(   curr_setting(Name, Module, Type, Default0, _Comment),
 	    eval_default(Default0, Module, Type, Default)
 	->  (   Value == Default
 	    ->	retract_setting(Module:Name)
@@ -326,6 +367,28 @@ restore_setting(QName) :-
 	    ;	true
 	    )
 	;   true
+	).
+
+%%	set_setting_default(:Name, +Default) is det.
+%
+%	Change the default for a setting.  The   effect  is  the same as
+%	set_setting/2, but the new value is  considered the default when
+%	saving and restoring  a  setting.  It   is  intended  to  change
+%	application defaults in a particular context.
+
+set_setting_default(QName, Default) :-
+	strip_module(QName, Module, Name),
+	must_be(atom, Name),
+	(   current_setting(Name, Module, Type, Default0, _Comment, _Src)
+	->  retractall(settings:default(Name, Module, _)),
+	    retract_setting(Module:Name),
+	    (   Default == Default0
+	    ->	true
+	    ;	assert(settings:default(Name, Module, Default))
+	    ),
+	    eval_default(Default, Module, Type, Value),
+	    set_setting(Module:Name, Value)
+	;   existence_error(setting, Module:Name)
 	).
 
 
@@ -395,17 +458,19 @@ read_setting(In, Term) :-
 %	Store setting loaded from file in the Prolog database.
 
 store_setting(setting(Module:Name, Value), _) :-
-	current_setting(Name, Module, Type, Default0, _Commentm, _Src), !,
+	curr_setting(Name, Module, Type, Default0, _Commentm), !,
 	eval_default(Default0, Module, Type, Default),
 	(   Value == Default
 	->  true
-	;   check_type(Type, Value),
+	;   check_type(Type, Value)
+	->  retractall(value(Name, Module, _)),
 	    assert(value(Name, Module, Value)),
 	    broadcast(settings(changed(Module:Name, Default, Value)))
 	).
-store_setting(setting(Module:Name, Value), Options) :-
+store_setting(setting(Module:Name, Value), Options) :- !,
 	(   option(undefined(load), Options, load)
-	->  assert(value(Name, Module, Value))
+	->  retractall(value(Name, Module, _)),
+	    assert(value(Name, Module, Value))
 	;   existence_error(setting, Module:Name)
 	).
 store_setting(Term, _) :-
@@ -442,7 +507,7 @@ write_setting_header(Out) :-
 	format(Out, '*/~n~n', []).
 
 save_setting(Out, Module:Name) :-
-	current_setting(Name, Module, Type, Default, Comment, _Src),
+	curr_setting(Name, Module, Type, Default, Comment),
 	(   value(Name, Module, Value),
 	    \+ ( eval_default(Default, Module, Type, DefValue),
 		 debug(setting, '~w <-> ~w~n', [DefValue, Value]),
@@ -479,11 +544,11 @@ current_setting(Module:Name) :-
 setting_property(Setting, Property) :-
 	ground(Setting), !,
 	Setting = Module:Name,
-	current_setting(Name, Module, Type, Default, Comment, _Src), !,
+	curr_setting(Name, Module, Type, Default, Comment), !,
 	setting_property(Property, Module, Type, Default, Comment).
 setting_property(Setting, Property) :-
 	Setting = Module:Name,
-	current_setting(Name, Module, Type, Default, Comment, _Src),
+	curr_setting(Name, Module, Type, Default, Comment),
 	setting_property(Property, Module, Type, Default, Comment).
 
 setting_property(type(Type),       _, Type, _,        _).
@@ -503,7 +568,7 @@ list_settings :-
 	       list_setting(Setting)).
 
 list_setting(Module:Name) :-
-	current_setting(Name, Module, Type, Default0, Comment, _Src),
+	curr_setting(Name, Module, Type, Default0, Comment),
 	eval_default(Default0, Module, Type, Default),
 	setting(Module:Name, Value),
 	(   Value \== Default
