@@ -1,4 +1,4 @@
-/*  $Id: pl-alloc.c,v 1.129 2007/06/19 09:05:30 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -643,7 +643,8 @@ outOfStack(void *stack, stack_overflow_action how)
     case STACK_OVERFLOW_RAISE:
     { fid_t fid = PL_open_foreign_frame();
       LD->outofstack = NULL;
-      gc_status.requested = FALSE;	/* can't have that */
+      PL_clearsig(SIG_GC);
+      s->gced_size = 0;			/* after handling, all is new */
       PL_unify_term(LD->exception.tmp,
 		    PL_FUNCTOR, FUNCTOR_error2,
 		      PL_FUNCTOR, FUNCTOR_resource_error1,
