@@ -1433,8 +1433,7 @@ typedef enum
   CHP_FOREIGN,				/* Foreign code choicepoint */
   CHP_TOP,				/* First (toplevel) choice */
   CHP_CATCH,				/* $catch initiated choice */
-  CHP_DEBUG,				/* Enable redo */
-  CHP_NONE				/* not a choice any-more */
+  CHP_DEBUG				/* Enable redo */
 } choice_type;
 
 typedef enum
@@ -1517,7 +1516,10 @@ struct queryFrame
   Word	       *aSave;			/* saved argument-stack */
   int		solutions;		/* # of solutions produced */
   Choice	saved_bfr;		/* Saved choice-point */
+  Code		saved_PC;		/* Saved program pointer */
+#ifndef LIFE_GC
   QueryFrame	parent;			/* Only valid inside GC */
+#endif
   struct choice	choice;			/* First (dummy) choice-point */
   LocalFrame	saved_environment;	/* Parent local-frame */
 					/* Do not put anything between */
@@ -1808,7 +1810,8 @@ Defining built-in predicates using the new interface
 		 *******************************/
 
 #if HAVE_SIGNAL
-#define MAXSIGNAL		32	/* highest system signal number */
+#define MAXSIGNAL		64	/* highest system signal number */
+#define SIG_PROLOG_OFFSET	32	/* Start of Prolog signals */
 
 typedef RETSIGTYPE (*handler_t)(int);
 typedef void *SignalContext;		/* struct sigcontext on sun */
@@ -1821,9 +1824,15 @@ typedef struct
 } sig_handler, *SigHandler;
 #endif /* HAVE_SIGNAL */
 
-#define SIG_EXCEPTION	   9		/* cannot be caught anyway */
-#define SIG_ATOM_GC	  30		/* `safe' and reserved signals */
-#define SIG_THREAD_SIGNAL 31
+#define SIG_EXCEPTION	  (SIG_PROLOG_OFFSET+0)
+#ifdef O_ATOMGC
+#define SIG_ATOM_GC	  (SIG_PROLOG_OFFSET+1)
+#endif
+#define SIG_GC		  (SIG_PROLOG_OFFSET+2)
+#ifdef O_PLMT
+#define SIG_THREAD_SIGNAL (SIG_PROLOG_OFFSET+3)
+#endif
+
 
 		 /*******************************
 		 *	   OPTION LISTS		*
@@ -2031,12 +2040,13 @@ typedef struct redir_context
 		*********************************/
 
 typedef struct
-{ bool		requested;		/* GC is requested by stack expander */
-  int		blocked;		/* GC is blocked now */
+{ int		blocked;		/* GC is blocked now */
   bool		active;			/* Currently running? */
-  intptr_t		collections;		/* # garbage collections */
+  long		collections;		/* # garbage collections */
   int64_t	global_gained;		/* global stack bytes collected */
   int64_t	trail_gained;		/* trail stack bytes collected */
+  int64_t	global_left;		/* global stack bytes left after GC */
+  int64_t	trail_left;		/* trail stack bytes left after GC */
   real		time;			/* time spent in collections */
 } pl_gc_status_t;
 
