@@ -261,6 +261,15 @@ static struct signame
 #ifdef SIGPWR
   { SIGPWR,	"pwr",    0},
 #endif
+  { SIG_EXCEPTION,     "prolog:exception",     0 },
+#ifdef SIG_ATOM_GC
+  { SIG_ATOM_GC,   "prolog:atom_gc",       0 },
+#endif
+  { SIG_GC,	       "prolog:gc",	       0 },
+#ifdef SIG_THREAD_SIGNAL
+  { SIG_THREAD_SIGNAL, "prolog:thread_signal", 0 },
+#endif
+
   { -1,		NULL,     0}
 };
 
@@ -280,6 +289,13 @@ signal_name(int sig)
 static int
 signal_index(const char *name)
 { struct signame *sn = signames;
+  char tmp[12];
+
+  if ( strncmp(name, "SIG", 3) == 0 && strlen(name) < 12 )
+  { strcpy(tmp, name+3);
+    strlwr(tmp);
+    name = tmp;
+  }
 
   for( ; sn->name; sn++ )
   { if ( streq(sn->name, name) )
@@ -291,20 +307,24 @@ signal_index(const char *name)
 
 
 int
-_PL_get_signum(term_t sig, int *n)
+PL_get_signum_ex(term_t sig, int *n)
 { char *s;
   int i = -1;
 
-  if ( !PL_get_integer(sig, &i) )
-  { if ( PL_get_atom_chars(sig, &s) )
-      i = signal_index(s);
+  if ( PL_get_integer(sig, &i) )
+  { 
+  } else if ( PL_get_chars(sig, &s, CVT_ATOM) )
+  { i = signal_index(s);
+  } else
+  { return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_signal, sig);
   }
+
   if ( i > 0 && i < 32 )		/* where to get these? */
   { *n = i;
     return TRUE;
   }
 
-  return FALSE;
+  return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_signal, sig);
 }
 
 
@@ -341,7 +361,7 @@ to the main thread.
 #define LD LOCAL_LD
 
 static void
-dispatch_signal(int sig, int sync)
+dispatch_signal(int sig, int sync, Code PC)
 { GET_LD
   SigHandler sh = &GD->sig_handlers[sig];
   fid_t fid;
@@ -380,7 +400,7 @@ dispatch_signal(int sig, int sync)
 	sysError("Recursively received fatal signal %d", sig);
   }
 
-  if ( gc_status.active )
+  if ( gc_status.active && sig < SIG_PROLOG_OFFSET )
   { fatalError("Received signal %d (%s) while in %ld-th garbage collection",
 	       sig, signal_name(sig), gc_status.collections);
   }
@@ -390,7 +410,8 @@ dispatch_signal(int sig, int sync)
     return;
   }
 
-  blockGC(PASS_LD1);
+  if ( !sync )
+    blockGC(PASS_LD1);
   LD->current_signal = sig;
   LD->sync_signal = sync;
   fid = PL_open_signal_foreign_frame();
@@ -408,13 +429,20 @@ dispatch_signal(int sig, int sync)
 			PL_Q_CATCH_EXCEPTION,
 			sh->predicate,
 			sigterm);
+    if ( PC )
+      QueryFromQid(qid)->saved_PC = PC;
     if ( !PL_next_solution(qid) && (except = PL_exception(qid)) )
     { PL_cut_query(qid);
-      unblockGC(PASS_LD1);
+      if ( !sync )
+	unblockGC(PASS_LD1);
       PL_throw(except);
       return;				/* make sure! */
     } else
-      PL_close_query(qid);
+    { if ( sync )
+	PL_cut_query(qid);
+      else
+	PL_close_query(qid);
+    }
   } else if ( true(sh, PLSIG_THROW) )
   { char *predname;
     int  arity;
@@ -428,7 +456,8 @@ dispatch_signal(int sig, int sync)
     }
       
     PL_error(predname, arity, NULL, ERR_SIGNALLED, sig, signal_name(sig));
-    unblockGC(PASS_LD1);
+    if ( !sync )
+      unblockGC(PASS_LD1);
 
     PL_throw(exception_term);		/* throw longjmp's */
     return;				/* make sure! */
@@ -444,10 +473,14 @@ dispatch_signal(int sig, int sync)
 
   LD->current_signal = saved_current_signal;
   LD->sync_signal = saved_sync;
-  PL_discard_foreign_frame(fid);
+  if ( sync )
+    PL_close_foreign_frame(fid);
+  else
+    PL_discard_foreign_frame(fid);
   lTop = addPointer(lBase, lTopSave);
 
-  unblockGC(PASS_LD1);
+  if ( !sync )
+    unblockGC(PASS_LD1);
 }
 
 #undef LD
@@ -455,7 +488,7 @@ dispatch_signal(int sig, int sync)
 
 static void
 pl_signal_handler(int sig)
-{ dispatch_signal(sig, FALSE);
+{ dispatch_signal(sig, FALSE, NULL);
 }
 
 #ifndef SA_RESTART
@@ -528,7 +561,8 @@ prepareSignal(int sig)
 
   if ( false(sh, PLSIG_PREPARED) )
   { set(sh, PLSIG_PREPARED);
-    sh->saved_handler = set_sighandler(sig, pl_signal_handler);
+    if ( sig < SIG_PROLOG_OFFSET )
+      sh->saved_handler = set_sighandler(sig, pl_signal_handler);
   }
 
   return sh;
@@ -555,6 +589,37 @@ hupHandler(int sig)
 }
 
 
+static void
+sig_exception_handler(int sig)
+{ if ( LD && LD->pending_exception )
+  { record_t ex = LD->pending_exception;
+	  
+    LD->pending_exception = 0;
+
+    PL_put_variable(exception_bin);
+    PL_recorded(ex, exception_bin);
+    PL_erase(ex);
+    exception_term = exception_bin;
+
+    SECURE(checkData(valTermRef(exception_term)));
+  }
+}
+
+
+static void
+agc_handler(int sig)
+{ if ( GD->statistics.atoms >= GD->atoms.non_garbage + GD->atoms.margin &&
+       !gc_status.blocked )
+    pl_garbage_collect_atoms();
+}
+
+
+static void
+gc_handler(int sig)
+{ garbageCollect(NULL, NULL);
+}
+
+
 static void
 initSignals(void)
 { struct signame *sn = signames;
@@ -570,6 +635,15 @@ initSignals(void)
     }
   }
 
+  PL_signal(SIG_EXCEPTION|PL_SIGSYNC, sig_exception_handler);
+  PL_signal(SIG_GC|PL_SIGSYNC, gc_handler);
+
+#ifdef SIG_THREAD_SIGNAL
+  PL_signal(SIG_THREAD_SIGNAL|PL_SIGSYNC, executeThreadSignals);
+#endif
+#ifdef SIG_ATOM_GC
+  PL_signal(SIG_ATOM_GC|PL_SIGSYNC, agc_handler);
+#endif
 #ifdef SIGHUP
   PL_signal(SIGHUP, hupHandler);
 #endif
@@ -736,6 +810,17 @@ PL_signal(int sigandflags, handler_t func)
 
 int
 PL_handle_signals()
+{ PL_local_data_t *ld = LD;
+
+  if ( !ld || ld->critical || !ld->pending_signals )
+    return 0;
+
+  return handleSignals(NULL);
+}
+
+
+int
+handleSignals(Code PC)
 { int done = 0;
   PL_local_data_t *ld = LD;
 
@@ -743,7 +828,7 @@ PL_handle_signals()
     return 0;
 
   while(ld->pending_signals)
-  { uintptr_t mask = 1;
+  { int64_t mask = 1;
     int sig = 1;
 
     for( ; mask ; mask <<= 1, sig++ )
@@ -751,47 +836,18 @@ PL_handle_signals()
       { ld->pending_signals &= ~mask;	/* reset the signal */
 
 	done++;
-
-#ifdef O_PLMT
-        if ( sig == SIG_THREAD_SIGNAL )
-	  executeThreadSignals(sig);
-	else
-#endif
-#ifdef O_ATOMGC
-	if ( sig == SIG_ATOM_GC )
-	{ if ( GD->statistics.atoms >= GD->atoms.non_garbage + GD->atoms.margin &&
-	       !gc_status.blocked )
-	    pl_garbage_collect_atoms();
-	} else
-#endif
-        if ( sig == SIG_EXCEPTION && ld->pending_exception )
-	{ record_t ex = ld->pending_exception;
-	  
-	  ld->pending_exception = 0;
-
-	  PL_put_variable(exception_bin);
-	  PL_recorded(ex, exception_bin);
-	  PL_erase(ex);
-	  exception_term = exception_bin;
-
-	  SECURE(checkData(valTermRef(exception_term)));
-	} else
-	  dispatch_signal(sig, TRUE);
+	dispatch_signal(sig, TRUE, PC);
 
 	if ( exception_term )
-	  return -1;
+	  goto out;
       }
     }
   }
 
+out:
   if ( exception_term )
     return -1;
 
-  if ( gc_status.requested )
-  { DEBUG(1, Sprintf("GC from PL_handle_signals()\n"));
-    garbageCollect(NULL, NULL);
-  }
-
   return done;
 }
 
@@ -988,6 +1044,7 @@ emptyStacks()
 #ifdef O_ATTVAR
   LD->attvar.head	= PL_new_term_ref();
   LD->attvar.tail       = PL_new_term_ref();
+  DEBUG(3, Sdprintf("attvar.tail at %p\n", valTermRef(LD->attvar.tail)));
 #endif
 #ifdef O_GVAR
   destroyGlobalVars();
